// Copyright (C) 2024 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:wyatt_continuous_deployment/src/core/enums/flag.dart';
import 'package:wyatt_continuous_deployment/src/core/injection/injectable.dart';
import 'package:wyatt_continuous_deployment/src/core/task/task_cubit.dart';
import 'package:wyatt_continuous_deployment/src/domain/entities/config/init_config_parameters.dart';
import 'package:wyatt_continuous_deployment/src/domain/entities/context.dart';
import 'package:wyatt_continuous_deployment/src/domain/usecases/config/init_config_usecase.dart';

part 'init_project_state.dart';

class InitProjectTask extends TaskCubit {
  factory InitProjectTask() => InitProjectTask._(
        initConfigUsecase: getIt(),
        checkToolsUsecase: getIt(),
        checkConfigFileUsecase: getIt(),
      );

  InitProjectTask._({
    required InitConfigUsecase initConfigUsecase,
    required super.checkToolsUsecase,
    required super.checkConfigFileUsecase,
  }) : _initConfigUsecase = initConfigUsecase;

  final InitConfigUsecase _initConfigUsecase;

  @override
  bool get checks => false; // No need to check config file (not created yet)

  @override
  Future<void> onRun(Context ctx) async {
    // 1. Init config
    emit(
      InitProjectState(
        path: ctx.configFile.path,
        force: ctx.flags.contains(Flag.force),
      ),
    );
    final initResult = await _initConfigUsecase(
      InitConfigParameters(
        path: ctx.configFile.path,
        force: ctx.flags.contains(Flag.force),
      ),
    );

    if (initResult.isErr) {
      return emit(TaskStateError.fromRes(initResult));
    }

    // 2. Check tools
    emit(TaskStateCheckTools());

    final checkToolsResult = await checkToolsUsecase(null);
    if (checkToolsResult.isErr) {
      return emit(TaskStateError.fromRes(checkToolsResult));
    }

    // 3. Check config file
    emit(TaskStateCheckConfigFile());

    final checkConfigFileResult =
        await checkConfigFileUsecase(ctx.configFile.path);
    if (checkConfigFileResult.isErr) {
      return emit(TaskStateError.fromRes(checkConfigFileResult));
    }

    return emit(const TaskStateSucceed(message: 'Project initialized'));
  }
}
