// Copyright (C) 2024 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';
import 'dart:io';

import 'package:wyatt_architecture/wyatt_architecture.dart';
import 'package:wyatt_continuous_deployment/src/core/utils/logging.dart';
import 'package:wyatt_continuous_deployment/src/domain/usecases/usecase.dart';
import 'package:wyatt_type_utils/wyatt_type_utils.dart';

abstract class ProcessUsecase<T> extends AsyncUseCase<T, void> {
  const ProcessUsecase();

  /// If true, the process will not throw an exception if the exit
  /// code is not 0.
  /// This is useful for commands that may return a non-zero exit code.
  /// For example, `git diff` returns 1 if there are differences.
  bool get ignoreOnFailure => false;

  String get onStartedMessage;

  String get onCompletedMessage;

  String get onErrorMessage;

  @override
  FutureOr<void> onStart(T? params) {
    if (params == null) {
      throw const ClientException('ProcessUsecase: You must provide a param');
    }
  }

  @override
  FutureOrResult<void> execute(T? params) async {
    final progress = logger.progress(onStartedMessage);
    final processResult = await run(params as T);
    if (processResult != null) {
      if (processResult.exitCode != 0) {
        progress.fail(onErrorMessage);
      }
      if (processResult.exitCode != 0 && !ignoreOnFailure) {
        throw ClientException(processResult.stderr.toString());
      }
    }
    progress.complete(onCompletedMessage);
    logger.detail(processResult?.stdout.toString().trim());
    return const Ok(null);
  }

  /// Run the process with the given parameters.
  Future<ProcessResult?> run(T param);
}

abstract class ProcessNoParamsUsecase extends NoParamsAsyncUseCase<void> {
  const ProcessNoParamsUsecase() : super();

  /// If true, the process will not throw an exception if the exit
  /// code is not 0.
  /// This is useful for commands that may return a non-zero exit code.
  /// For example, `git diff` returns 1 if there are differences.
  bool get ignoreOnFailure => false;

  String get onStartedMessage;

  String get onCompletedMessage;

  String get onErrorMessage;

  @override
  FutureOrResult<void> execute(void params) async {
    final progress = logger.progress(onStartedMessage);
    final processResult = await run();
    if (processResult != null) {
      if (processResult.exitCode != 0) {
        progress.fail(onErrorMessage);
      }
      if (processResult.exitCode != 0 && !ignoreOnFailure) {
        throw ClientException(processResult.stderr.toString());
      }
    }
    progress.complete(onCompletedMessage);
    logger.detail(processResult?.stdout.toString().trim());
    return const Ok(null);
  }

  /// Run the process with no parameters.
  Future<ProcessResult?> run();
}
