// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:wyatt_crud_bloc/src/features/crud/blocs/crud_base_cubit/crud_base_cubit.dart';

/// {@template crud_builder}
/// A widget that builds itself based on the latest snapshot of interaction
/// with a [CrudBaseCubit].
/// {@endtemplate}
class CrudBuilder<
    Bloc extends StateStreamable<CrudState>,
    Initial extends CrudState,
    Loading extends CrudState,
    Success extends CrudState,
    Error extends CrudState> extends StatelessWidget {
  /// {@macro crud_builder}
  const CrudBuilder({
    required this.builder,
    required this.initialBuilder,
    required this.loadingBuilder,
    required this.errorBuilder,
    this.unknownBuilder,
    this.buildWhen,
    super.key,
  });

  /// {@macro crud_builder}
  ///
  /// This factory constructor is used to create a [CrudBuilder] with
  /// [CrudState]s. `S` is the Success State, and it must be a subtype of
  /// [CrudSuccess]. It the only type that you have to specify.
  static CrudBuilder<Bloc, CrudInitial, CrudLoading, CrudSuccess, CrudError>
      typed<Bloc extends StateStreamable<CrudState>,
              Success extends CrudSuccess>({
    required Widget Function(BuildContext, Success) builder,
    required Widget Function(BuildContext, CrudInitial) initialBuilder,
    required Widget Function(BuildContext, CrudLoading) loadingBuilder,
    required Widget Function(BuildContext, CrudError) errorBuilder,
    Widget Function(BuildContext, CrudState)? unknownBuilder,
    BlocBuilderCondition<CrudState>? buildWhen,
  }) =>
          CrudBuilder<Bloc, CrudInitial, CrudLoading, Success, CrudError>(
            builder: builder,
            initialBuilder: initialBuilder,
            loadingBuilder: loadingBuilder,
            errorBuilder: errorBuilder,
            unknownBuilder: unknownBuilder,
            buildWhen: buildWhen,
          );

  /// {@macro crud_builder}
  ///
  /// This factory constructor is used to create a [CrudBuilder] wich reacts
  /// only to [CrudLoading] states.
  static CrudBuilder<Bloc, CrudInitial, CrudLoading, CrudSuccess, CrudError>
      onLoading<Bloc extends StateStreamable<CrudState>>({
    required Widget? Function(BuildContext, CrudLoading) builder,
    Widget Function(BuildContext, CrudState)? otherBuilder,
    BlocBuilderCondition<CrudState>? buildWhen,
  }) =>
          CrudBuilder<Bloc, CrudInitial, CrudLoading, CrudSuccess, CrudError>(
            loadingBuilder: (context, state) =>
                builder.call(context, state) ??
                otherBuilder?.call(context, state) ??
                const SizedBox.shrink(),
            builder: (context, state) =>
                otherBuilder?.call(context, state) ?? const SizedBox.shrink(),
            initialBuilder: (context, state) =>
                otherBuilder?.call(context, state) ?? const SizedBox.shrink(),
            errorBuilder: (context, state) =>
                otherBuilder?.call(context, state) ?? const SizedBox.shrink(),
            unknownBuilder: otherBuilder,
            buildWhen: buildWhen,
          );

  /// {@macro crud_builder}
  ///
  /// This factory constructor is used to create a [CrudBuilder] wich reacts
  /// only to [CrudInitial] states.
  static CrudBuilder<Bloc, CrudInitial, CrudLoading, CrudSuccess, CrudError>
      onInitial<Bloc extends StateStreamable<CrudState>>({
    required Widget? Function(BuildContext, CrudInitial) builder,
    Widget Function(BuildContext, CrudState)? otherBuilder,
    BlocBuilderCondition<CrudState>? buildWhen,
  }) =>
          CrudBuilder<Bloc, CrudInitial, CrudLoading, CrudSuccess, CrudError>(
            initialBuilder: (context, state) =>
                builder.call(context, state) ??
                otherBuilder?.call(context, state) ??
                const SizedBox.shrink(),
            builder: (context, state) =>
                otherBuilder?.call(context, state) ?? const SizedBox.shrink(),
            loadingBuilder: (context, state) =>
                otherBuilder?.call(context, state) ?? const SizedBox.shrink(),
            errorBuilder: (context, state) =>
                otherBuilder?.call(context, state) ?? const SizedBox.shrink(),
            unknownBuilder: otherBuilder,
            buildWhen: buildWhen,
          );

  /// {@macro crud_builder}
  ///
  /// This factory constructor is used to create a [CrudBuilder] wich reacts
  /// only to [CrudError] states.
  static CrudBuilder<Bloc, CrudInitial, CrudLoading, CrudSuccess, CrudError>
      onError<Bloc extends StateStreamable<CrudState>>({
    required Widget? Function(BuildContext, CrudError) builder,
    Widget Function(BuildContext, CrudState)? otherBuilder,
    BlocBuilderCondition<CrudState>? buildWhen,
  }) =>
          CrudBuilder<Bloc, CrudInitial, CrudLoading, CrudSuccess, CrudError>(
            errorBuilder: (context, state) =>
                builder.call(context, state) ??
                otherBuilder?.call(context, state) ??
                const SizedBox.shrink(),
            builder: (context, state) =>
                otherBuilder?.call(context, state) ?? const SizedBox.shrink(),
            loadingBuilder: (context, state) =>
                otherBuilder?.call(context, state) ?? const SizedBox.shrink(),
            initialBuilder: (context, state) =>
                otherBuilder?.call(context, state) ?? const SizedBox.shrink(),
            unknownBuilder: otherBuilder,
            buildWhen: buildWhen,
          );

  final BlocBuilderCondition<CrudState>? buildWhen;
  final Widget Function(BuildContext context, Success state) builder;
  final Widget Function(BuildContext context, Initial state) initialBuilder;
  final Widget Function(BuildContext context, Loading state) loadingBuilder;
  final Widget Function(BuildContext context, Error state) errorBuilder;
  final Widget Function(BuildContext context, CrudState state)? unknownBuilder;

  @override
  Widget build(BuildContext context) => BlocBuilder<Bloc, CrudState>(
        buildWhen: buildWhen,
        builder: (context, state) => switch (state) {
          Success() => builder(context, state),
          Initial() => initialBuilder(context, state),
          Loading() => loadingBuilder(context, state),
          Error() => errorBuilder(context, state),
          _ => unknownBuilder?.call(context, state) ??
              Center(
                child: Text(
                  'Unknown state: $state',
                ),
              ),
        },
      );
}
